/* Microprocessors Based Design
 * Asynchronous Web Server
 *
 * Calin Cruceru <calin.cruceru@stud.acs.upb.ro>, 335CB
 *
 * Utility functions which use the http_parser lib.
 */

#include "http_util.h"
#include "debug.h"

#include <http_parser.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>


/* Need to store them at global scope as they are used in the callback
 * function.
 */
static char *request_path_ptr;
static http_parser request_parser;

static int on_url_cb(http_parser *parser, const char *buf, size_t len)
{
	assert(parser == &request_parser);

	memcpy(request_path_ptr, buf, len);
	request_path_ptr[len] = '\0';

	return 0;
}

/* Use mostly null settings except for on_path callback */
static const http_parser_settings settings_on_path = {
	0,		/* on_message_begin */
	on_url_cb,	/* on_url */
	0,		/* on_status */
	0,		/* on_header_field */
	0,		/* on_header_value */
	0,		/* on_headers_complete */
	0,		/* on_body */
	0,		/* on_message_complete */
	0,		/* on_chunk_header */
	0,		/* on_chunk_complete */
};

void extract_path(const char *http_req, char *out_buffer)
{
	size_t bytes_parsed;

	/* init HTTP_REQUEST parser */
	request_path_ptr = out_buffer;
	http_parser_init(&request_parser, HTTP_REQUEST);

	bytes_parsed = http_parser_execute(&request_parser,
					   &settings_on_path,
					   http_req,
					   strlen(http_req));

	dlog(LOG_DEBUG, "Parsed HTTP request (bytes: %zu), path: %s\n",
			bytes_parsed, out_buffer);
}

