/* Proiectare cu Microprocesoare
 * True Random Number Generator
 *
 * 2016, Calin Cruceru <calin.cruceru@stud.acs.upb.ro> (335CB)
 *
 * Entry point in the random number generator program
 */
#include <stdio.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>

#define HIGH_LOW_THRESHOLD 480

static void ADC_init(void)
{
	ADMUX = 0;
	ADMUX |= _BV(MUX0) | _BV(MUX1) | _BV(MUX2); /* set pin 7 */
	ADMUX |= _BV(REFS0); /* AVCC with external capacitor at AREF pin */

	ADCSRA = 0;
	ADCSRA |= _BV(ADPS1); /* prescaler at 4 */
	ADCSRA |= _BV(ADIE); /* enable interrupt */
	ADCSRA |= _BV(ADEN); /* enable ADC */
}

ISR(ADC_vect)
{
	static unsigned char high; /* 1 if it's currently high, 0 otherwise */
	static unsigned int samples;
	static unsigned char byte;
	static unsigned char bits_ready;

	unsigned int adc_val = ADC;

	samples++;

	if (high == 0 && adc_val > HIGH_LOW_THRESHOLD)
		 high = 1;

	if (high == 1 && adc_val < HIGH_LOW_THRESHOLD) {
		byte = (byte << 1) + samples % 2;

		if (++bits_ready == 8) {
			fwrite(&byte, 1, 1, stdout);
			byte = 0;
			bits_ready = 0;
		}

		high = 0;
		samples = 0;
	}

	/* Start a new conversion */
	ADCSRA |= _BV(ADSC);
}

int main(void)
{
	/* Initialize ADC */
	ADC_init();

	/* Enable interrupts */
	sei();

	/* Start a new conversion */
	ADCSRA |= _BV(ADSC);

	/* Wait forever */
	for (;;) {
		;
	}

	return 0;
}

