/* Proiectare cu Microprocesoare
 * True Random Number Generator
 *
 * 2016, Calin Cruceru <calin.cruceru@stud.acs.upb.ro> (335CB)
 *
 * The module which initializes standard I/O through UART.
 */
#include <stdio.h>
#include <avr/io.h>

static int uart_putchar(char c, __attribute__((unused)) FILE *stream)
{
	loop_until_bit_is_set(UCSR0A, UDRE0);
	UDR0 = c;

	return 0;
}

static int uart_getchar(__attribute__((unused)) FILE *stream)
{
	loop_until_bit_is_set(UCSR0A, RXC0);

	return UDR0;
}

static void uart_init(void)
{
	/*
	 * Setting baud rate.
	 */
#define BAUD 9600
#include <util/setbaud.h>
	UBRR0H = UBRRH_VALUE;
	UBRR0L = UBRRL_VALUE;
#if USE_2X
	UCSR0A |= (1 << U2X0);
#else
	UCSR0A &= ~(1 << U2X0);
#endif
	/*
	 * Default frame format: 8 data bits, 1 stop bit, no parity.
	 * Enable Rx/Tx.
	 */
	UCSR0B = (1<<TXEN0) | (1<<RXEN0);
}

__attribute__((constructor))
static void init_uart_io(void)
{
	uart_init();
	fdevopen(uart_putchar, NULL);
	fdevopen(NULL, uart_getchar);
}

